@@ -1,7 +1,7 @@
 /*
- *  WindowMaker window manager
+ *  Window Maker window manager
  * 
- *  Copyright (c) 1997 Alfredo K. Kojima
+ *  Copyright (c) 1997-2002 Alfredo K. Kojima
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,8 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ *  USA.
  */
 
 #ifndef WMCLIENT_H_
@@ -31,7 +32,9 @@ void wClientKill(WWindow *wwin);
 void wClientCheckProperty(WWindow *wwin, XPropertyEvent *event);
 
 
-void GetNormalHints(WWindow *wwin, XWindowAttributes *wattribs, int geometry);
+void wClientGetNormalHints(WWindow *wwin, XWindowAttributes *wattribs,
+			   Bool geometry, int *x, int *y, unsigned *width,
+			   unsigned *height);
 void GetColormapWindows(WWindow *wwin);
 
 #endif
