@@ -2,7 +2,8 @@
  * 
  *  Window Maker window manager
  * 
- *  Copyright (c) 1997, 1998 Alfredo K. Kojima
+ *  Copyright (c) 1997-2002 Alfredo K. Kojima
+ *  Copyright (c) 1998-2002 Dan Pascu
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,7 +23,8 @@
 
 #include "wconfig.h"
 
-#include <assert.h>
+#ifndef LITE
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -50,16 +52,16 @@
 #include "framewin.h"
 #include "session.h"
 #include "xmodifier.h"
-#include <proplist.h>
 
-#include "list.h"
+#include <WINGs/WUtil.h>
+
 
 
 extern char *Locale;
 
 extern WDDomain *WDRootMenu;
 
-extern wCursor[WCUR_LAST];
+extern Cursor wCursor[WCUR_LAST];
 
 extern Time LastTimestamp;
 
@@ -111,6 +113,7 @@ static Shortcut *shortcutList = NULL;
  * REFRESH - forces the desktop to be repainted
  * EXIT [QUICK] - exit the window manager [without confirmation]
  * EXEC <program> - execute an external program
+ * SHEXEC <command> - execute a shell command
  * WORKSPACE_MENU - places the workspace submenu
  * ARRANGE_ICONS
  * RESTART [<window manager>] - restarts the window manager
@@ -125,6 +128,7 @@ static Shortcut *shortcutList = NULL;
  *                the resulting menu in current position. The output of
  *                command must be a valid menu description.
  *                The space between '|' and command is optional.
+ *                || will do the same, but will not cache the contents.
  * SAVE_SESSION - saves the current state of the desktop, which include
  *		  all running applications, all their hints (geometry,
  *		  position on screen, workspace they live on, the dock
@@ -135,15 +139,13 @@ static Shortcut *shortcutList = NULL;
  *		  CLEAR_SESSION is used. If SaveSessionOnExit = Yes; in
  *		  WindowMaker domain file, then saving is automatically
  *		  done on every windowmaker exit, overwriting any
- *		  SAVE_SESSION or CLEAR_SESSION (see below).
+ *		  SAVE_SESSION or CLEAR_SESSION (see below). Also save
+ *		  dock state now.
  * CLEAR_SESSION - clears any previous saved session. This will not have
  *		  any effect if SaveSessionOnExit is True.
  *							     
  */
 
-#define MAX(a,b)	((a)>(b) ? (a) : (b))
-
-
 #define M_QUICK		1
 
 /* menu commands */
@@ -152,19 +154,6 @@ static void
 execCommand(WMenu *menu, WMenuEntry *entry)
 {
     char *cmdline;
-    static char *shell = NULL;
-
-    /*
-     * This have a problem: if the shell is tcsh (not sure about others)
-     * and ~/.tcshrc have /bin/stty erase ^H somewhere on it, the shell
-     * will block and the command will not be executed.
-    if (!shell) {
-	shell = getenv("SHELL");
-	if (!shell)
-	    shell = "/bin/sh";
-    }
-     */
-    shell = "/bin/sh";
 
     cmdline = ExpandOptions(menu->frame->screen_ptr, (char*)entry->clientdata);
 
@@ -174,27 +163,8 @@ execCommand(WMenu *menu, WMenuEntry *entry)
     XSync(dpy, 0);
 
     if (cmdline) {
-        /* We should not use ParseCommand() here, because it does not handle
-         * correctly '~/' expansion and '>' redirection.
-         * While we can do '~/' expansion ourselves, we can do nothing about
-         * '>' redirection.
-         * So we better let /bin/sh to do that for us. Dan.
-	 * Ok. -Alfredo
-         */
-        if (fork()==0) {
-
-	    SetupEnvironment(menu->frame->screen_ptr);
-
-	    CloseDescriptors();
-
-#ifdef HAVE_SETPGID
-            setpgid(0, 0);
-#endif
-            execl(shell, shell, "-c", cmdline, NULL);
-	    wsyserror("could not exec %s -c %s\n", shell, cmdline);
-            Exit(-1);
-        }
-	free(cmdline);
+	ExecuteShellCommand(menu->frame->screen_ptr, cmdline);
+	wfree(cmdline);
     }
     XUngrabPointer(dpy, CurrentTime);
     XSync(dpy, 0);
@@ -205,33 +175,44 @@ static void
 exitCommand(WMenu *menu, WMenuEntry *entry)
 {
     static int inside = 0;
+    int result;
 
     /* prevent reentrant calls */
     if (inside)
 	return;
     inside = 1;
 
-    if ((int)entry->clientdata==M_QUICK
-	|| wMessageDialog(menu->frame->screen_ptr, _("Exit"),
-			  _("Exit window manager?"), 
-			  _("Exit"), _("Cancel"), NULL)==WAPRDefault) {
-	int i;
+#define R_CANCEL 0
+#define R_EXIT   1
+
+    result = R_CANCEL;
+
+    if ((long)entry->clientdata==M_QUICK) {
+        result = R_EXIT;
+    } else {
+        int r, oldSaveSessionFlag;
+
+        oldSaveSessionFlag = wPreferences.save_session_on_exit;
+        r = wExitDialog(menu->frame->screen_ptr, _("Exit"),
+                        _("Exit window manager?"),
+                        _("Exit"), _("Cancel"), NULL);
+
+        if (r==WAPRDefault) {
+            result = R_EXIT;
+        } else if (r==WAPRAlternate) {
+            /* Don't modify the "save session on exit" flag if the
+             * user canceled the operation. */
+            wPreferences.save_session_on_exit = oldSaveSessionFlag;
+        }
+    }
+    if (result==R_EXIT) {
 #ifdef DEBUG
 	printf("Exiting WindowMaker.\n");
 #endif
-	for (i=0; i<wScreenCount; i++) {
-	    WScreen *scr;
-
-	    scr = wScreenWithNumber(i);
-	    if (scr)
-		wScreenSaveState(scr);
-	}
-
-	RestoreDesktop(NULL);
-	
-	ExecExitScript();
-	Exit(0);
+	Shutdown(WSExitMode);
     }
+#undef R_EXIT
+#undef R_CANCEL
     inside = 0;
 }
 
@@ -241,7 +222,6 @@ shutdownCommand(WMenu *menu, WMenuEntry *entry)
 {
     static int inside = 0;
     int result;
-    int i;
 
     /* prevent reentrant calls */
     if (inside)
@@ -254,10 +234,10 @@ shutdownCommand(WMenu *menu, WMenuEntry *entry)
 
     
     result = R_CANCEL;
-    if ((int)entry->clientdata==M_QUICK)
+    if ((long)entry->clientdata==M_QUICK)
 	result = R_CLOSE;
     else {
-#ifdef R6SM
+#ifdef XSMP_ENABLED
 	if (wSessionIsManaged()) {
 	    int r;
 	    
@@ -273,36 +253,33 @@ shutdownCommand(WMenu *menu, WMenuEntry *entry)
 	} else
 #endif
 	{
-	    int r;
+	    int r, oldSaveSessionFlag;
 
-	    r = wMessageDialog(menu->frame->screen_ptr, 
-			       _("Kill X session"), 
-			       _("Kill Window System session?\n"
-				 "(all applications will be closed)"),
-			       _("Kill"), _("Cancel"), NULL);
-	    if (r==WAPRDefault)
-		result = R_KILL;
+            oldSaveSessionFlag = wPreferences.save_session_on_exit;
+
+	    r = wExitDialog(menu->frame->screen_ptr,
+                            _("Kill X session"),
+                            _("Kill Window System session?\n"
+                              "(all applications will be closed)"),
+                            _("Kill"), _("Cancel"), NULL);
+            if (r==WAPRDefault) {
+                result = R_KILL;
+            } else if (r==WAPRAlternate) {
+                /* Don't modify the "save session on exit" flag if the
+                 * user canceled the operation. */
+                wPreferences.save_session_on_exit = oldSaveSessionFlag;
+            }
 	}
     } 
     
     if (result!=R_CANCEL) {
-#ifdef R6SM
+#ifdef XSMP_ENABLED
 	if (result == R_CLOSE) {
-	    wSessionRequestShutdown();
+	    Shutdown(WSLogoutMode);
 	} else 
-#endif /* R6SM */
+#endif /* XSMP_ENABLED */
 	{
-	    for (i=0; i<wScreenCount; i++) {
-		WScreen *scr;
-		
-		scr = wScreenWithNumber(i);
-		if (scr)
-		    wScreenSaveState(scr);
-	    }
-	    WipeDesktop(NULL);
-	    
-	    ExecExitScript();
-	    Exit(0);
+	    Shutdown(WSKillMode);
 	}
     }
 #undef R_CLOSE
@@ -315,17 +292,9 @@ shutdownCommand(WMenu *menu, WMenuEntry *entry)
 static void
 restartCommand(WMenu *menu, WMenuEntry *entry)
 {
-    int i;
-    
-    for (i=0; i<wScreenCount; i++) {
-	WScreen *scr;
-	
-	scr = wScreenWithNumber(i);
-	if (scr)
-	    wScreenSaveState(scr);
-    }
-    RestoreDesktop(NULL);
-    Restart((char*)entry->clientdata);
+    Shutdown(WSRestartPreparationMode);
+    Restart((char*)entry->clientdata, False);
+    Restart(NULL, True);
 }
 
 
@@ -358,7 +327,10 @@ hideOthersCommand(WMenu *menu, WMenuEntry *entry)
 static void
 saveSessionCommand(WMenu *menu, WMenuEntry *entry)
 {
-    wSessionSaveState(menu->frame->screen_ptr);
+    if (!wPreferences.save_session_on_exit)
+	wSessionSaveState(menu->frame->screen_ptr);
+
+    wScreenSaveState(menu->frame->screen_ptr);
 }
 
 
@@ -366,6 +338,7 @@ static void
 clearSessionCommand(WMenu *menu, WMenuEntry *entry)
 {
     wSessionClearState(menu->frame->screen_ptr);
+    wScreenSaveState(menu->frame->screen_ptr);
 }
 
 
@@ -383,9 +356,48 @@ legalPanelCommand(WMenu *menu, WMenuEntry *entry)
 }
 
 
-
 /********************************************************************/
 
+
+static char*
+getLocalizedMenuFile(char *menu)
+{
+    char *buffer;
+    char *ptr;
+    int len;
+
+    if (!Locale)
+	return NULL;
+
+    len = strlen(menu)+strlen(Locale)+8;
+    buffer = wmalloc(len);
+
+    /* try menu.locale_name */
+    snprintf(buffer, len, "%s.%s", menu, Locale);
+    if (access(buffer, F_OK)==0) {
+	return buffer;
+    }
+    /* check if it is in the form aa_bb.encoding and check for aa_bb */
+    ptr = strchr(Locale, '.');
+    if (ptr) {
+	*ptr = 0;
+	if (access(buffer, F_OK)==0) {
+	    return buffer;
+	}
+    }
+    /* now check for aa */
+    ptr = strchr(buffer, '_');
+    if (ptr) {
+	*ptr = 0;
+	if (access(buffer, F_OK)==0) {
+	    return buffer;
+	}
+    }
+
+    return NULL;
+}
+
+
 static void
 raiseMenus(WMenu *menu)
 {
@@ -405,22 +417,24 @@ raiseMenus(WMenu *menu)
 Bool
 wRootMenuPerformShortcut(XEvent *event)
 {
+    WScreen *scr = wScreenForRootWindow(event->xkey.root);
     Shortcut *ptr;
     int modifiers;
     int done = 0;
-    
+
     /* ignore CapsLock */
     modifiers = event->xkey.state & ValidModMask;
-    
+
     for (ptr = shortcutList; ptr!=NULL; ptr = ptr->next) {
-	if (ptr->keycode==0) 
-	    continue;
-	
-	if (ptr->keycode==event->xkey.keycode && (ptr->modifier==modifiers)) {
-	    (*ptr->entry->callback)(ptr->menu, ptr->entry);
-	    done = True;
-	}
+        if (ptr->keycode==0 || ptr->menu->menu->screen_ptr!=scr)
+            continue;
+
+        if (ptr->keycode==event->xkey.keycode && ptr->modifier==modifiers) {
+            (*ptr->entry->callback)(ptr->menu, ptr->entry);
+            done = True;
+        }
     }
+
     return done;
 }
 
@@ -457,7 +471,7 @@ rebindKeygrabs(WScreen *scr)
     while (wwin!=NULL) {
 	XUngrabKey(dpy, AnyKey, AnyModifier, wwin->frame->core->window);
 
-	if (!wwin->window_flags.no_bind_keys) {
+	if (!WFLAGP(wwin, no_bind_keys)) {
 	    wWindowSetKeyGrabs(wwin);
 	}
 	wwin = wwin->prev;
@@ -475,7 +489,7 @@ removeShortcutsForMenu(WMenu *menu)
     while (ptr!=NULL) {
 	tmp = ptr->next;
 	if (ptr->menu == menu) {
-	    free(ptr);
+	    wfree(ptr);
 	} else {
 	    ptr->next = newList;
 	    newList = ptr;
@@ -509,9 +523,8 @@ addShortcut(char *file, char *shortcutDefinition, WMenu *menu,
 	*k = 0;
 	mod = wXModifierFromKey(b);
 	if (mod<0) {
-	    wwarning(_("%s:invalid key modifier \"%s\""), file,
-		     shortcutDefinition);
-	    free(ptr);
+	    wwarning(_("%s:invalid key modifier \"%s\""), file, b);
+	    wfree(ptr);
 	    return False;
 	}
 	ptr->modifier |= mod;
@@ -525,7 +538,7 @@ addShortcut(char *file, char *shortcutDefinition, WMenu *menu,
     if (ksym==NoSymbol) {
 	wwarning(_("%s:invalid kbd shortcut specification \"%s\" for entry %s"),
 		 file, shortcutDefinition, entry->text);
-	free(ptr);
+	wfree(ptr);
 	return False;
     }
     
@@ -533,7 +546,7 @@ addShortcut(char *file, char *shortcutDefinition, WMenu *menu,
     if (ptr->keycode==0) {
 	wwarning(_("%s:invalid key in shortcut \"%s\" for entry %s"), file,
 		 shortcutDefinition, entry->text);
-	free(ptr);
+	wfree(ptr);
 	return False;
     }
 
@@ -622,7 +635,7 @@ static void
 separateCommand(char *line, char ***file, char **command)
 {
     char *token, *tmp = line;
-    LinkedList *list = NULL;
+    WMArray *array = WMCreateArray(4);
     int count, i;
 
     *file = NULL;
@@ -637,20 +650,19 @@ separateCommand(char *line, char ***file, char **command)
                     wwarning(_("%s: missing command"), line);
                 break;
             }
-            list = list_cons(token, list);
+	    WMAddToArray(array, token);
         }
     } while (token!=NULL && tmp!=NULL);
 
-    count = list_length(list);
+    count = WMGetArrayItemCount(array);
     if (count>0) {
         *file = wmalloc(sizeof(char*)*(count+1));
-        i = count;
         (*file)[count] = NULL;
-        while (list!=NULL) {
-            (*file)[--i] = list->head;
-            list_remove_head(&list);
+        for (i = 0; i < count; i++) {
+            (*file)[i] = WMGetFromArray(array, i);
         }
     }
+    WMFreeArray(array);
 }
 
 
@@ -666,24 +678,27 @@ constructMenu(WMenu *menu, WMenuEntry *entry)
     time_t last=0;
     
     separateCommand((char*)entry->clientdata, &path, &cmd);
-    if (!path || *path==NULL || **path==0) {
+    if (path == NULL || *path==NULL || **path==0) {
 	wwarning(_("invalid OPEN_MENU specification: %s"),
 		 (char*)entry->clientdata);
 	return;
     }
-    
+
     if (path[0][0]=='|') {
         /* pipe menu */
-	
+
         if (!menu->cascades[entry->cascade] || 
             menu->cascades[entry->cascade]->timestamp == 0) { 
             /* parse pipe */
 
 	    submenu = readMenuPipe(menu->frame->screen_ptr, path);
 
-            /* there's no automatic reloading */
-            if(submenu != NULL)
-                submenu->timestamp = 1;
+            if(submenu != NULL) {
+		if (path[0][1] == '|')
+		    submenu->timestamp = 0;
+		else
+                    submenu->timestamp = 1; /* there's no automatic reloading */
+	    }
         } else { 
             submenu = NULL; 
         }
@@ -693,38 +708,21 @@ constructMenu(WMenu *menu, WMenuEntry *entry)
         while(path[i] != NULL) {
 	    char *tmp;
 
+	    if (strcmp(path[i], "-noext")==0) {
+		i++;
+		continue;
+	    }
+	    
 	    tmp = wexpandpath(path[i]);
-	    free(path[i]);
-	    path[i] = tmp;
-
-            if (Locale) {
-                lpath = wmalloc(strlen(path[i])+32);
-
-                strcpy(lpath, path[i]);
-                strcat(lpath, ".");
-                strcat(lpath, Locale);
-                if (stat(lpath, &stat_buf)<0) {
-                    int i;
-                    i = strlen(Locale);
-                    if (i>2) {
-                        lpath[strlen(lpath)-(i-2)]=0;
-                        if (stat(lpath, &stat_buf)==0) {
-                            free(path[i]);
-                            path[i] = lpath;
-                            lpath = NULL;
-                        }
-                    }
-                } else {
-                    free(path[i]);
-                    path[i] = lpath;
-                    lpath = NULL;
-                }
-            }
-
-            if (lpath) {
-                free(lpath);
-                lpath = NULL;
-            }
+	    wfree(path[i]);
+	    lpath = getLocalizedMenuFile(tmp);
+	    if (lpath) {
+		wfree(tmp);
+		path[i] = lpath;
+		lpath = NULL;
+	    } else {
+		path[i] = tmp;
+	    }
 
             if (stat(path[i], &stat_buf)==0) {
                 if (last < stat_buf.st_mtime)
@@ -740,7 +738,7 @@ constructMenu(WMenu *menu, WMenuEntry *entry)
         }
 
         if (first < 0) {
-            wsyserror(_("%s:could not stat menu :%s"), "OPEN_MENU",
+            wsyserror(_("%s:could not stat menu:%s"), "OPEN_MENU",
                       (char*)entry->clientdata);
             goto finish;
         }
@@ -751,7 +749,7 @@ constructMenu(WMenu *menu, WMenuEntry *entry)
 	    if (S_ISDIR(stat_buf.st_mode)) {
 		/* menu directory */
 		submenu = readMenuDirectory(menu->frame->screen_ptr,
-                                            entry->text, &path[first], cmd);
+                                            entry->text, path, cmd);
 		if (submenu)
 		    submenu->timestamp = last;
 	    } else if (S_ISREG(stat_buf.st_mode)) {
@@ -780,10 +778,18 @@ constructMenu(WMenu *menu, WMenuEntry *entry)
 finish:
     i = 0;
     while (path[i]!=NULL)
-        free(path[i++]);
-    free(path);
+        wfree(path[i++]);
+    wfree(path);
     if (cmd)
-	free(cmd);
+	wfree(cmd);
+}
+
+
+static void
+cleanupWorkspaceMenu(WMenu *menu)
+{
+    if (menu->frame->screen_ptr->workspace_menu == menu)
+	menu->frame->screen_ptr->workspace_menu = NULL;
 }
 
 
@@ -798,10 +804,12 @@ addWorkspaceMenu(WScreen *scr, WMenu *menu, char *title)
 	return NULL;
     } else {
 	scr->flags.added_workspace_menu = 1;
-	
+
 	wsmenu = wWorkspaceMenuMake(scr, True);
+	wsmenu->on_destroy = cleanupWorkspaceMenu;
+
 	scr->workspace_menu = wsmenu;
-	entry=wMenuAddCallback(menu, title, NULL, NULL);
+	entry = wMenuAddCallback(menu, title, NULL, NULL);
 	wMenuEntrySetCascade(menu, entry, wsmenu);
 	
 	wWorkspaceMenuUpdate(scr, wsmenu);
@@ -810,6 +818,43 @@ addWorkspaceMenu(WScreen *scr, WMenu *menu, char *title)
 }
 
 
+static void
+cleanupWindowsMenu(WMenu *menu)
+{
+    if (menu->frame->screen_ptr->switch_menu == menu)
+	menu->frame->screen_ptr->switch_menu = NULL;
+}
+
+
+static WMenuEntry*
+addWindowsMenu(WScreen *scr, WMenu *menu, char *title)
+{
+    WMenu *wwmenu;
+    WWindow *wwin;
+    WMenuEntry *entry;
+
+    if (scr->flags.added_windows_menu) {
+	wwarning(_("There are more than one WINDOWS_MENU commands in the applications menu. Only one is allowed."));
+	return NULL;
+    } else {
+	scr->flags.added_windows_menu = 1;
+
+	wwmenu = wMenuCreate(scr, _("Window List"), False);
+	wwmenu->on_destroy = cleanupWindowsMenu;
+	scr->switch_menu = wwmenu;
+	wwin = scr->focused_window;
+	while (wwin) {
+	    UpdateSwitchMenu(scr, wwin, ACTION_ADD);
+
+	    wwin = wwin->prev;
+	}
+	entry = wMenuAddCallback(menu, title, NULL, NULL);
+	wMenuEntrySetCascade(menu, entry, wwmenu);
+    }
+    return entry;
+}
+
+
 static WMenuEntry*
 addMenuEntry(WMenu *menu, char *title, char *shortcut, char *command,
 	     char *params, char *file_name)
@@ -844,7 +889,18 @@ addMenuEntry(WMenu *menu, char *title, char *shortcut, char *command,
 	    wwarning(_("%s:missing parameter for menu command \"%s\""),
 		     file_name, command);
 	else {
-	    entry = wMenuAddCallback(menu, title, execCommand, wstrdup(params));
+	    entry = wMenuAddCallback(menu, title, execCommand, 
+				     wstrconcat("exec ", params));
+	    entry->free_cdata = free;
+	    shortcutOk = True;
+	}
+    } else if (strcmp(command, "SHEXEC")==0) {
+	if (!params)
+	    wwarning(_("%s:missing parameter for menu command \"%s\""),
+		     file_name, command);
+	else {
+	    entry = wMenuAddCallback(menu, title, execCommand,
+				     wstrdup(params));
 	    entry->free_cdata = free;
 	    shortcutOk = True;
 	}
@@ -872,6 +928,10 @@ addMenuEntry(WMenu *menu, char *title, char *shortcut, char *command,
     } else if (strcmp(command, "WORKSPACE_MENU")==0) {
 	entry = addWorkspaceMenu(scr, menu, title);
 	
+	shortcutOk = True;
+    } else if (strcmp(command, "WINDOWS_MENU")==0) {
+	entry = addWindowsMenu(scr, menu, title);
+
 	shortcutOk = True;
     } else if (strcmp(command, "ARRANGE_ICONS")==0) {
 	entry = wMenuAddCallback(menu, title, arrangeIconsCommand, NULL);
@@ -1132,8 +1192,9 @@ readMenuFile(WScreen *scr, char *file_name)
 	if (!args) {
 	    wwarning(_("could not make arguments for menu file preprocessor"));
 	} else {
-	    sprintf(command, "%s %s %s", CPP_PATH, args, file_name);
-	    free(args);
+	    snprintf(command, sizeof(command), "%s %s %s", 
+		     CPP_PATH, args, file_name);
+	    wfree(args);
 	    file = popen(command, "r");
 	    if (!file) {
 		wsyserror(_("%s:could not open/preprocess menu file"), 
@@ -1209,7 +1270,7 @@ readMenuPipe(WScreen *scr, char **file_name)
     char params[MAXLINE];
     char shortcut[MAXLINE];
     char *line;
-    char * filename;
+    char *filename;
     char flat_file[MAXLINE];
     int i;
 #ifdef USECPP
@@ -1219,11 +1280,12 @@ readMenuPipe(WScreen *scr, char **file_name)
 
     flat_file[0] = '\0';
 
-    for(i = 0 ; file_name[i] != NULL ; i++) {
+    for(i=0; file_name[i]!=NULL; i++) {
 	strcat(flat_file, file_name[i]);
 	strcat(flat_file, " ");
     }
-    filename = flat_file+1;
+    filename = flat_file + (flat_file[1]=='|'?2:1);
+
 
 #ifdef USECPP
     if (!wPreferences.flags.nocpp) {
@@ -1231,9 +1293,10 @@ readMenuPipe(WScreen *scr, char **file_name)
        if (!args) {
            wwarning(_("could not make arguments for menu file preprocessor"));
        } else {
-           sprintf(command, "%s | %s %s", filename, CPP_PATH, args);
+           snprintf(command, sizeof(command), "%s | %s %s", 
+		    filename, CPP_PATH, args);
 
-           free(args);
+           wfree(args);
            file = popen(command, "r");
            if (!file) {
                wsyserror(_("%s:could not open/preprocess menu file"), filename);
@@ -1295,13 +1358,35 @@ typedef struct {
 
 
 static int
-myCompare(dir_data *d1, dir_data *d2)
+myCompare(const void *d1, const void *d2)
 {
-    return strcmp(d1->name, d2->name);
+    dir_data *p1 = *(dir_data**)d1;
+    dir_data *p2 = *(dir_data**)d2;
+
+    return strcmp(p1->name, p2->name);
 }
 
 
 /************  Menu Configuration From Directory   *************/
+
+static Bool
+isFilePackage(char *file)
+{
+    int l;
+
+    /* check if the extension indicates this file is a 
+     * file package. For now, only recognize .themed */
+
+    l = strlen(file);
+
+    if (l > 7 && strcmp(&(file[l-7]), ".themed")==0) {
+	return True;
+    } else {
+	return False;
+    }
+}
+
+
 static WMenu*
 readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
 {
@@ -1310,12 +1395,24 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
     struct stat stat_buf;
     WMenu *menu=NULL;
     char *buffer;
-    LinkedList *dirs = NULL, *files = NULL;
+    WMArray *dirs = NULL, *files = NULL;
+    WMArrayIterator iter;
     int length, i, have_space=0;
     dir_data *data;
+    int stripExtension = 0;
 
+    
+    dirs = WMCreateArray(16);
+    files = WMCreateArray(16);
+    
     i=0;
     while (path[i]!=NULL) {
+	if (strcmp(path[i], "-noext")==0) {
+	    stripExtension = 1;
+	    i++;
+	    continue;
+	}
+
         dir = opendir(path[i]);
         if (!dir) {
             i++;
@@ -1331,7 +1428,7 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
 	    if (dentry->d_name[0] == '.')
 		continue;
 
-            buffer = wmalloc(strlen(path[i])+strlen(dentry->d_name)+4);
+            buffer = malloc(strlen(path[i])+strlen(dentry->d_name)+4);
             if (!buffer) {
                 wsyserror(_("out of memory while constructing directory menu %s"),
                          path[i]);
@@ -1346,72 +1443,84 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
                 wsyserror(_("%s:could not stat file \"%s\" in menu directory"),
                           path[i], dentry->d_name);
             } else {
-                data = (dir_data*) wmalloc(sizeof(dir_data));
-                data->name = wstrdup(dentry->d_name);
-                data->index = i;
-                if (S_ISDIR(stat_buf.st_mode)) {
+		Bool isFilePack = False;
+
+		data = NULL;
+                if (S_ISDIR(stat_buf.st_mode)
+		    && !(isFilePack = isFilePackage(dentry->d_name))) {
+
                     /* access always returns success for user root */
                     if (access(buffer, X_OK)==0) {
                         /* Directory is accesible. Add to directory list */
-                        list_insert_sorted(data, &dirs, (int(*)())myCompare);
-                        data = NULL;
+
+			data = (dir_data*) wmalloc(sizeof(dir_data));
+			data->name = wstrdup(dentry->d_name);
+			data->index = i;
+
+			WMAddToArray(dirs, data);
                     }
-                } else if (S_ISREG(stat_buf.st_mode)) {
+                } else if (S_ISREG(stat_buf.st_mode) || isFilePack) {
                     /* Hack because access always returns X_OK success for user root */
 #define S_IXANY (S_IXUSR | S_IXGRP | S_IXOTH)
                     if ((command!=NULL && access(buffer, R_OK)==0) ||
                         (command==NULL && access(buffer, X_OK)==0 &&
                          (stat_buf.st_mode & S_IXANY))) {
-                        list_insert_sorted(data, &files, (int(*)())myCompare);
-                        data = NULL;
+
+			data = (dir_data*) wmalloc(sizeof(dir_data));
+			data->name = wstrdup(dentry->d_name);
+			data->index = i;
+
+			WMAddToArray(files, data);
                     }
                 }
-                if (data!=NULL) {
-                    if (data->name)
-                        free(data->name);
-                    free(data);
-                    data = NULL;
-                }
             }
-            free(buffer);
+            wfree(buffer);
         }
 
         closedir(dir);
         i++;
     }
-
-    if (!dirs && !files)
+    
+    if (!WMGetArrayItemCount(dirs) && !WMGetArrayItemCount(files)) {
+	WMFreeArray(dirs);
+	WMFreeArray(files);
         return NULL;
+    }
+
+    WMSortArray(dirs, myCompare);
+    WMSortArray(files, myCompare);
 
     menu = wMenuCreate(scr, title, False);
     menu->on_destroy = removeShortcutsForMenu;
-    
-    while (dirs != NULL) {
+
+    WM_ITERATE_ARRAY(dirs, data, iter) {
         /* New directory. Use same OPEN_MENU command that was used
          * for the current directory. */
-        dir_data *d = (dir_data*)dirs->head;
-
-        length = strlen(path[d->index])+strlen(d->name)+6;
+        length = strlen(path[data->index])+strlen(data->name)+6;
+	if (stripExtension)
+	    length += 7;
         if (command)
             length += strlen(command) + 6;
-        buffer = wmalloc(length);
+        buffer = malloc(length);
         if (!buffer) {
             wsyserror(_("out of memory while constructing directory menu %s"),
-                     path[d->index]);
+                     path[data->index]);
             break;
         }
+	
+	buffer[0] = '\0';
+        if (stripExtension)
+            strcat(buffer, "-noext ");
+
+        have_space = strchr(path[data->index], ' ')!=NULL ||
+                     strchr(data->name, ' ')!=NULL;
+
+	if (have_space)
+            strcat(buffer, "\"");
+	strcat(buffer, path[data->index]);
 
-        have_space = strchr(path[d->index], ' ')!=NULL ||
-                     strchr(d->name, ' ')!=NULL;
-        if (have_space) {
-            buffer[0] = '"';
-            buffer[1] = 0;
-            strcat(buffer, path[d->index]);
-        } else {
-            strcpy(buffer, path[d->index]);
-        }
         strcat(buffer, "/");
-        strcat(buffer, d->name);
+        strcat(buffer, data->name);
         if (have_space)
             strcat(buffer, "\"");
         if (command) {
@@ -1419,64 +1528,64 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
             strcat(buffer, command);
         }
 
-        addMenuEntry(menu, d->name, NULL, "OPEN_MENU", buffer, path[d->index]);
+        addMenuEntry(menu, data->name, NULL, "OPEN_MENU", buffer, path[data->index]);
 
-        free(buffer);
-        if (dirs->head) {
-            if (d->name)
-                free(d->name);
-            free(dirs->head);
-        }
-        list_remove_head(&dirs);
+        wfree(buffer);
+	if (data->name)
+	    wfree(data->name);
+	wfree(data);
     }
 
-    while (files != NULL) {
+    WM_ITERATE_ARRAY(files, data, iter) {
         /* executable: add as entry */
-        dir_data *f = (dir_data*) files->head;;
-
-        length = strlen(path[f->index])+strlen(f->name)+6;
+        length = strlen(path[data->index])+strlen(data->name)+6;
         if (command)
             length += strlen(command);
 
-        buffer = wmalloc(length);
+        buffer = malloc(length);
         if (!buffer) {
             wsyserror(_("out of memory while constructing directory menu %s"),
-                     path[f->index]);
+                     path[data->index]);
             break;
         }
 
-        have_space = strchr(path[f->index], ' ')!=NULL ||
-                     strchr(f->name, ' ')!=NULL;
+        have_space = strchr(path[data->index], ' ')!=NULL ||
+                     strchr(data->name, ' ')!=NULL;
         if (command!=NULL) {
             strcpy(buffer, command);
             strcat(buffer, " ");
             if (have_space)
                 strcat(buffer, "\"");
-            strcat(buffer, path[f->index]);
+            strcat(buffer, path[data->index]);
         } else {
             if (have_space) {
                 buffer[0] = '"';
                 buffer[1] = 0;
-                strcat(buffer, path[f->index]);
+                strcat(buffer, path[data->index]);
             } else {
-                strcpy(buffer, path[f->index]);
+                strcpy(buffer, path[data->index]);
             }
         }
         strcat(buffer, "/");
-        strcat(buffer, f->name);
+        strcat(buffer, data->name);
         if (have_space)
             strcat(buffer, "\"");
 
-        addMenuEntry(menu, f->name, NULL, "EXEC", buffer, path[f->index]);
+	if (stripExtension) {
+	    char *ptr = strrchr(data->name, '.');
+	    if (ptr && ptr!=data->name)
+		*ptr = 0;
+	}
+        addMenuEntry(menu, data->name, NULL, "SHEXEC", buffer, path[data->index]);
 
-        free(buffer);
-        if (files->head) {
-            if (f->name)
-                free(f->name);
-            free(files->head);
-        }
-        list_remove_head(&files);
+        wfree(buffer);
+	if (data->name)
+	    wfree(data->name);
+	wfree(data);
     }
+    
+    WMFreeArray(files);
+    WMFreeArray(dirs);
 
     return menu;
 }
@@ -1491,6 +1600,8 @@ makeDefaultMenu(WScreen *scr)
     
     menu = wMenuCreate(scr, _("Commands"), True);
     wMenuAddCallback(menu, "XTerm", execCommand, "xterm");
+    wMenuAddCallback(menu, "rxvt", execCommand, "rxvt");
+    wMenuAddCallback(menu, _("Restart"), restartCommand, NULL);
     wMenuAddCallback(menu, _("Exit..."), exitCommand, NULL);
     return menu;
 }
@@ -1507,53 +1618,25 @@ makeDefaultMenu(WScreen *scr)
  *---------------------------------------------------------------------- 
  */
 static WMenu*
-configureMenu(WScreen *scr, proplist_t definition)
+configureMenu(WScreen *scr, WMPropList *definition)
 {
     WMenu *menu = NULL;
-    proplist_t elem;
+    WMPropList *elem;
     int i, count;
-    proplist_t title, command, params;
+    WMPropList *title, *command, *params;
     char *tmp, *mtitle;
 
     
-    if (PLIsString(definition)) {
+    if (WMIsPLString(definition)) {
 	struct stat stat_buf;
 	char *path = NULL;
 	Bool menu_is_default = False;
 
 	/* menu definition is a string. Probably a path, so parse the file */
-	
-	tmp = wexpandpath(PLGetString(definition));
 
-	if (Locale) {
-	    path = wmalloc(strlen(tmp)+32);
+	tmp = wexpandpath(WMGetFromPLString(definition));
 
-	    strcpy(path, tmp);
-	    strcat(path, ".");
-	    strcat(path, Locale);
-
-	    /* look for menu.xy */
-	    if (stat(path, &stat_buf)<0) {		
-		int i;
-		i = strlen(Locale);
-		if (i>2) {
-		    path[strlen(path)-(i-2)]=0;
-		    /* look for menu.xy_zw */
-		    if (stat(path, &stat_buf)<0) {
-			free(path);
-			/* If did not find any localized menus, try
-			 * only menu. This can also mean that
-			 * the path in WMRootMenu was already the
-			 * path for the localized menu (eg: menu = "menu.ab")
-			 */
-			path = NULL;
-		    }
-		} else {
-		    free(path);
-		    path = NULL;
-		}
-	    }
-	}
+	path = getLocalizedMenuFile(tmp);
 
 	if (!path)
 	    path = wfindfile(DEF_CONFIG_PATHS, tmp);
@@ -1566,14 +1649,14 @@ configureMenu(WScreen *scr, proplist_t definition)
 	if (!path) {
 	    wsyserror(_("could not find menu file \"%s\" referenced in WMRootMenu"),
 		     tmp);
-	    free(tmp);
+	    wfree(tmp);
 	    return NULL;
 	}
-	
+
 	if (stat(path, &stat_buf)<0) {
 	    wsyserror(_("could not access menu \"%s\" referenced in WMRootMenu"), path);
-	    free(path);
-	    free(tmp);
+	    wfree(path);
+	    wfree(tmp);
 	    return NULL;
 	}
 
@@ -1588,40 +1671,61 @@ configureMenu(WScreen *scr, proplist_t definition)
 
 	    menu = readMenuFile(scr, path);
 	    if (menu)
-		menu->timestamp = MAX(stat_buf.st_mtime, WDRootMenu->timestamp);
+		menu->timestamp = WMAX(stat_buf.st_mtime, WDRootMenu->timestamp);
 	} else {
 	    menu = NULL;
 	}
-	free(path);
-	free(tmp);
+	wfree(path);
+	wfree(tmp);
 
 	return menu;
     }
 
-    count = PLGetNumberOfElements(definition);
+    count = WMGetPropListItemCount(definition);
     if (count==0)
 	return NULL;
 
-    elem = PLGetArrayElement(definition, 0);
-    if (!PLIsString(elem)) {
-	tmp = PLGetDescription(elem);
+    elem = WMGetFromPLArray(definition, 0);
+    if (!WMIsPLString(elem)) {
+	tmp = WMGetPropListDescription(elem, False);
 	wwarning(_("%s:format error in root menu configuration \"%s\""), 
 		 "WMRootMenu", tmp);
-	free(tmp);
+	wfree(tmp);
 	return NULL;
     }
-    mtitle = PLGetString(elem);
+    mtitle = WMGetFromPLString(elem);
     
     menu = wMenuCreate(scr, mtitle, False);
     menu->on_destroy = removeShortcutsForMenu;
 
-    for (i=1; i<count; i++) {
-	elem = PLGetArrayElement(definition, i);
+#ifdef GLOBAL_SUBMENU_FILE
+    {
+	WMenu *submenu;
+	WMenuEntry *mentry;
 
-	if (!PLIsArray(elem) || PLGetNumberOfElements(elem) < 2)
+	submenu = readMenuFile(scr, GLOBAL_SUBMENU_FILE);
+
+	if (submenu) {
+	    mentry = wMenuAddCallback(menu, submenu->frame->title, NULL, NULL);
+	    wMenuEntrySetCascade(menu, mentry, submenu);
+	}
+    }
+#endif
+
+    for (i=1; i<count; i++) {
+	elem = WMGetFromPLArray(definition, i);
+#if 0
+	if (WMIsPLString(elem)) {
+	    char *file;
+	    
+	    file = WMGetFromPLString(elem);
+	    
+	}
+#endif	
+	if (!WMIsPLArray(elem) || WMGetPropListItemCount(elem) < 2)
 	    goto error;
 	
-	if (PLIsArray(PLGetArrayElement(elem,1))) {
+	if (WMIsPLArray(WMGetFromPLArray(elem,1))) {
 	    WMenu *submenu;
 	    WMenuEntry *mentry;
 	    
@@ -1634,35 +1738,35 @@ configureMenu(WScreen *scr, proplist_t definition)
 	    }
 	} else {
 	    int idx = 0;
-	    char *shortcut;
+	    WMPropList *shortcut;
 	    /* normal entry */
 
-	    title = PLGetArrayElement(elem, idx++);
-	    shortcut = PLGetArrayElement(elem, idx++);
-	    if (strcmp(PLGetString(shortcut), "SHORTCUT")==0) {
-		shortcut = PLGetArrayElement(elem, idx++);
-		command = PLGetArrayElement(elem, idx++);
+	    title = WMGetFromPLArray(elem, idx++);
+	    shortcut = WMGetFromPLArray(elem, idx++);
+	    if (strcmp(WMGetFromPLString(shortcut), "SHORTCUT")==0) {
+		shortcut = WMGetFromPLArray(elem, idx++);
+		command = WMGetFromPLArray(elem, idx++);
 	    } else {
 		command = shortcut;
 		shortcut = NULL;
 	    }
-	    params = PLGetArrayElement(elem, idx++);
+	    params = WMGetFromPLArray(elem, idx++);
 
 	    if (!title || !command)
 		goto error;
 	    
-	    addMenuEntry(menu, PLGetString(title), 
-			 shortcut ? PLGetString(shortcut) : NULL, 
-			 PLGetString(command),
-			 params ? PLGetString(params) : NULL, "WMRootMenu");
+	    addMenuEntry(menu, WMGetFromPLString(title), 
+			 shortcut ? WMGetFromPLString(shortcut) : NULL, 
+			 WMGetFromPLString(command),
+			 params ? WMGetFromPLString(params) : NULL, "WMRootMenu");
 	}
 	continue;
 
       error:
-	tmp = PLGetDescription(elem);
+	tmp = WMGetPropListDescription(elem, False);
 	wwarning(_("%s:format error in root menu configuration \"%s\""), 
 		 "WMRootMenu", tmp);
-	free(tmp);
+	wfree(tmp);
     }
     
     return menu;
@@ -1695,17 +1799,18 @@ void
 OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
 {
     WMenu *menu=NULL;
-    proplist_t definition;
+    WMPropList *definition;
     /*
-    static proplist_t domain=NULL;
+    static WMPropList *domain=NULL;
     
     if (!domain) {
-	domain = PLMakeString("WMRootMenu");
+	domain = WMCreatePLString("WMRootMenu");
     }
      */
     
     scr->flags.root_menu_changed_shortcuts = 0;
     scr->flags.added_workspace_menu = 0;
+    scr->flags.added_windows_menu = 0;
 
     if (scr->root_menu && scr->root_menu->flags.mapped) {
 	menu = scr->root_menu;
@@ -1717,8 +1822,7 @@ OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
 	    if (keyboard)
 		wMenuMapAt(menu, 0, 0, True);
 	    else
-		wMenuMapCopyAt(menu, x-menu->frame->core->width/2, 
-			       y-menu->frame->top_width/2);
+		wMenuMapCopyAt(menu, x-menu->frame->core->width/2, y);
 	}
 	return;
     }
@@ -1730,7 +1834,7 @@ OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
     definition = PLGetDomain(domain);
  */
     if (definition) {
-	if (PLIsArray(definition)) {
+	if (WMIsPLArray(definition)) {
 	    if (!scr->root_menu 
 		|| WDRootMenu->timestamp > scr->root_menu->timestamp) {
 		menu = configureMenu(scr, definition);
@@ -1747,22 +1851,40 @@ OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
     if (!menu) {
 	/* menu hasn't changed or could not be read */
 	if (!scr->root_menu) {
+	    wMessageDialog(scr, _("Error"),
+			  _("The applications menu could not be loaded. "
+			    "Look at the console output for a detailed "
+			    "description of the errors."),
+			   _("OK"), NULL, NULL);
+
 	    menu = makeDefaultMenu(scr);
 	    scr->root_menu = menu;
 	}
 	menu = scr->root_menu;
     } else {
 	/* new root menu */
-	if (scr->root_menu)
+	if (scr->root_menu) {
 	    wMenuDestroy(scr->root_menu, True);
+	}
 	scr->root_menu = menu;
     }
     if (menu) {
-	wMenuMapAt(menu, x-menu->frame->core->width/2, y-menu->frame->top_width/2,
-		   keyboard);
+        int newx, newy;
+
+        if (keyboard && x==0 && y==0) {
+            newx = newy = 0;
+        } else if (keyboard && x==scr->scr_width/2 && y==scr->scr_height/2) {
+            newx = x - menu->frame->core->width/2;
+            newy = y - menu->frame->core->height/2;
+        } else {
+            newx = x - menu->frame->core->width/2;
+            newy = y;
+        }
+        wMenuMapAt(menu, newx, newy, keyboard);
     }
     
     if (scr->flags.root_menu_changed_shortcuts)
 	rebindKeygrabs(scr);
 }
 
+#endif /* !LITE */
