@@ -2,7 +2,7 @@
  * 
  *  Raster graphics library
  * 
- *  Copyright (c) 1997 Alfredo K. Kojima
+ *  Copyright (c) 1997-2002 Alfredo K. Kojima
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -47,10 +47,6 @@ char *alloca ();
 
 #include <png.h>
 
-#if PNG_LIBPNG_VER < 96
-#error libpng_must_be_more_at_least_0_96
-#endif
-
 #include "wraster.h"
 
 
@@ -69,7 +65,7 @@ RLoadPNG(RContext *context, char *file, int index)
     png_uint_32 width, height;
     int depth, junk, color_type;
     png_bytep *png_rows;
-    unsigned char *r, *g, *b, *a;
+    unsigned char *ptr;
     
     f = fopen(file, "r");
     if (!f) {
@@ -105,7 +101,7 @@ RLoadPNG(RContext *context, char *file, int index)
 	fclose(f);
 	png_destroy_read_struct(&png, &pinfo, &einfo);
 	if (image)
-	  RDestroyImage(image);
+	  RReleaseImage(image);
 	return NULL;
     }
     
@@ -132,10 +128,10 @@ RLoadPNG(RContext *context, char *file, int index)
     }
     
     /* normalize to 8bpp with alpha channel */
-    if (color_type == PNG_COLOR_TYPE_PALETTE && depth < 8)
+    if (color_type == PNG_COLOR_TYPE_PALETTE && depth <= 8)
         png_set_expand(png);
 
-    if (color_type == PNG_COLOR_TYPE_GRAY && depth < 8)
+    if (color_type == PNG_COLOR_TYPE_GRAY && depth <= 8)
         png_set_expand(png);
 
     if (png_get_valid(png, pinfo, PNG_INFO_tRNS))
@@ -158,7 +154,8 @@ RLoadPNG(RContext *context, char *file, int index)
 	if (sgamma==0)
 	    sgamma = 1;
     } else {
-	sgamma = 2.0;
+	/* blah */
+	sgamma = 2.2;
     }
     
     if (png_get_gAMA(png, pinfo, &gamma))
@@ -180,7 +177,7 @@ RLoadPNG(RContext *context, char *file, int index)
     if (!png_rows) {
 	RErrorCode = RERR_NOMEMORY;
 	fclose(f);
-	RDestroyImage(image);
+	RReleaseImage(image);
 	png_destroy_read_struct(&png, &pinfo, &einfo);
 #ifdef C_ALLOCA
 	alloca(0);
@@ -192,7 +189,7 @@ RLoadPNG(RContext *context, char *file, int index)
 	if (!png_rows[y]) {
 	    RErrorCode = RERR_NOMEMORY;
 	    fclose(f);
-	    RDestroyImage(image);
+	    RReleaseImage(image);
 	    png_destroy_read_struct(&png, &pinfo, &einfo);
 #ifdef C_ALLOCA
 	    alloca(0);
@@ -209,27 +206,19 @@ RLoadPNG(RContext *context, char *file, int index)
     
     fclose(f);
 
-    r = image->data[0];
-    g = image->data[1];
-    b = image->data[2];
-    a = image->data[3];
+    ptr = image->data;
 
     /* convert to RImage */
     if (alpha) {
 	for (y=0; y<height; y++) {
-	    for (x=0, i=0; x<width; x++) {
-		*(r++) = *(png_rows[y]+i++);
-		*(g++) = *(png_rows[y]+i++);
-		*(b++) = *(png_rows[y]+i++);
-		*(a++) = *(png_rows[y]+i++);
+	    for (x=0, i=width*4; x<i; x++, ptr++) {
+		*ptr = *(png_rows[y]+x);
 	    }
 	}
     } else {
 	for (y=0; y<height; y++) {
-	    for (x=0, i=0; x<width; x++) {
-		*(r++) = *(png_rows[y]+i++);
-		*(g++) = *(png_rows[y]+i++);
-		*(b++) = *(png_rows[y]+i++);
+	    for (x=0, i=width*3; x<i; x++, ptr++) {
+		*ptr = *(png_rows[y]+x);
 	    }
 	}	
     }
